/*
 *   I2O Configuration utility
 *
 *   Common parts
 *
 *   (C) Copyright 1999 University of Helsinki,
 *                      Department of Computer Science
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version
 *   2 of the License, or (at your option) any later version.
 *
 *   Author     Taneli Vhkangas <Taneli.Vahakangas@cs.Helsinki.FI>
 *
 */


/*
 * CGI FORM handling part
 */
 
#include <ctype.h>
#include <stdarg.h>
/* For some odd reason it does not link without this. */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#undef _GNU_SOURCE
#include <sys/ioctl.h>
#include <unistd.h>

#include <linux/i2o-dev.h>

#include "common.h"

#define MAX_QUERY_SIZE (1<<12)

int querylen;
char *method;

/* This is called if we run out of memory or something
 * equivalently catastrophic happens.
 */

void html_error(const char *error)
{
	printf("Content-type: text/html\n\n");
	printf("<HTML><BODY>%s</BODY></HTML>\n", error);
	exit(0);
}

/*
 * Get the value part of variable=value type assignments
 * as are common in CGI forms. The assignments are separated
 * with '&' (ampersand) signs. If we encounter errors the
 * program is stopped. Most usually such errors come from
 * broken clients or are break-in attempts.
 *
 * IN: qstr = string to scan for assignments
 *     var = variable name to look for
 * OUT: value of given variable or NULL if not found
 */

char *extract(const char *qstr, const char *var)
{
	char *ptr, *str, *ret;

	if (!qstr || !var)
		return NULL;
	if (!(str=(char *)malloc(strlen(var)+2)))
		html_error("Out of memory");
	sprintf(str, "%s=", var);
	ptr=strstr(qstr, str);

	if (!ptr) return NULL;
	if (ptr==qstr || *(ptr-1)=='&') {
		ptr+=strlen(str);
		if (strchr(ptr, '&')) {
			if (!(ret=(char *)malloc(strchr(ptr, '&')-ptr+1)))
				html_error("Out of memory");
			strncpy(ret, ptr, strchr(ptr, '&')-ptr);
			ret[strchr(ptr, '&')-ptr]='\0';
		} else
			ret=strdup(ptr);
		free(str);
		return ret;
	}
	free(str);
	return NULL;
}

/*
 * Parse a CGI form string. The original string is preserved.
 *
 * IN: original string
 * OUT: parsed string
 */

char *decode(const char *cgistr)
{
	char *x = "0123456789ABCDEF";
	char *ret, *p, *q;
	char d;

	if (!cgistr) return NULL;
	if (!(ret=strdup(cgistr)))
		return NULL;
	p=ret; q=ret;
	while (*p) {
		switch (*p) {
		case '+':
			*q=' '; p++; break;
		case '%':
			if (*(p+1) && *(p+2))
				if (isxdigit(*(p+1)) && isxdigit(*(p+2))) {
					d=(strchr(x, toupper(*(p+1)))-x)*0x10+
						(strchr(x, toupper(*(p+2)))-x);
					*q=(char)d;
					p+=3;
				}
			break;
		default:
			*q=*p;
			p++;
		}
		q++;
	}
	*q='\0';
	return ret;
}

/*
 * Get query string. Works for POST, GET and PUT queries.
 *
 * IN: nothing
 * OUT: query string
 */

char *getquery(void)
{
	char *ret = NULL;

	method=getenv("REQUEST_METHOD");
	if (!method)
		html_error("Can't get REQUEST_METHOD");
	if (!strcmp(method, "POST") || !strcmp(method, "PUT")) {
		if ((querylen=atoi(getenv("CONTENT_LENGTH")))<1)
			querylen=MAX_QUERY_SIZE;
		if (!(ret=(char *)malloc(querylen+1)))
			html_error("Out of memory");
		if (!read(0, (void *)ret, querylen)) {
			free(ret);
			ret=NULL;
		}
	} else if (!strcmp(method, "GET"))
		ret=getenv("QUERY_STRING");
	return ret;
}





/*
 * HTML page handling
 */

/* Should we allocate more space dynamically? Currently we don't. */
#define PAGE_LEN 32768

/* XXX Move these inside the struct page */
char *prologue = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">";
char *epilogue = NULL;

FILE *output = NULL;
int offset = 0;
int outputfd = 1;

/*
 * Start a new page
 * IN: pointer to space allocated for a struct page
 * OUT: nothing
 *      in case of error value of p->outputfd == -1
 */

void start_page(struct page *p)
{
	if (!p) {
		p = (struct page *)malloc(sizeof(struct page));
		if (!p) {
			error_page("Can't allocate memory\n");
			exit(1);
		}
	}	

	p->outputfd=-1;

	p->output = stdout; /* This is a pretty good default */
	p->str = (char *)malloc(PAGE_LEN);
	if (!p->str)
		return;
	empty_page(p);
	strcpy(p->type, "text/html");
	p->binary = 0;
	p->offset = 0;
	p->outputfd = 1; /* This equals stdout */
}

/*
 * Print out an error page.
 * This can be used for short error messages if an early malloc (or similar)
 * fails and the user program can't call start_page.
 *
 * IN: output string
 * OUT: yes
 */
 
void error_page(const char *format, ...)
{
	va_list args;
	
	va_start(args, format);
	vprintf(format, args);
	va_end(args);
}

/*
 * Change MIME type for page.
 *
 * IN: pointer to page, new type as a string (may not exceed 40 characters)
 *
 */
 
void change_type(struct page *p, char *type)
{
	if (p && type) strncpy(p->type, type, strlen(type)>40?40:strlen(type));
}

/*
 * Append text to a page.
 *
 * IN: pointer to page
 *     printf-style string with variable arguments
 * OUT: zero for success
 *      non-zero for failure
 */
 
int hprintf(struct page *p, const char *format, ...)
{
	va_list args;
	char *s;
	int len;

	if (!(s = (char *)malloc(PAGE_LEN)))
		return 1;

	va_start(args, format);
	len = vsnprintf(p->str + p->offset, PAGE_LEN - p->offset, format, args);
	va_end(args);
	if (len==-1)
		return 1;
	p->offset += len;

	return 0;
}

/*
 * Paste binary data over the page.
 *
 * IN: pointer to page, pointer to data, data length
 *
 */

void write_page(struct page *p, char *data, unsigned int len)
{
	if (!data && !len)
	{
		p->str = NULL;
		p->binary = 0;
	}
	if (!p->binary)
	{
		p->binary = 1;
		free(p->str);
	}
	p->str = data;
	p->offset = len;
}

/*
 * Empty page. Make all hprintf'd or write_page'd stuff go away.
 *
 * IN: pointer to page
 */
 
void empty_page(struct page *p)
{
	if (!p) return;
	if (!p->binary && p->str) memset(p->str, 0, PAGE_LEN);
	else p->str = NULL;

	p->offset = 0;
}

/*
 * Send out the page and free resources allocated to it
 *
 * IN: pointer to page
 */

void end_page(struct page *p)
{
	int len = 0;
	
	if (!p->str)
		return;

	if (!p->binary)
	{
		if (prologue) len += strlen(prologue);
		if (epilogue) len += strlen(epilogue);
		len += p->offset;
		fprintf(p->output, "Content-length: %d\n"
				   "Content-type: %s\n\n", len, p->type);
	 	fprintf(p->output, "%s%s%s", prologue ? prologue : "", p->str,
						epilogue ? epilogue : "");
		free (p->str);
		p->offset = 0;
	}
	else 
	{
		fprintf(p->output, "Content-length: %d\n"
				   "Content-type: %s\n\n", p->offset, p->type);
		fflush(p->output);
		do
		{
			len = write(p->outputfd, p->str, p->offset);
			p->str += len;
			p->offset -= len;
		} while (len && p->offset);
	}
}






/*
 * I2O specific parts
 */



#define FRAGSIZE 8192
#define MAXLEN 8192

struct mod_info moduleinfo[NUM_MODULE_TYPES] = {
	{ "DDM",        0x01, "DDM" },
	{ "DDM_MPB",    0x02, "DDM Module Parameter Block" },
	{ "DDM_CDT",    0x03, "DDM Configuration Dialog Table" },
	{ "IRTOS",      0x11, "IRTOS" },
	{ "IRTOS_PRIV", 0x12, "IRTOS private module" },
	{ "IRTOS_CDT",  0x13, "IRTOS Dialogue Table" },
	{ "IOP_PRIV",   0x22, "IOP private module" },
	{ "IOP_CDT",    0x23, "IOP Dialogue Table" },
	{ "IOP_ARC",    0x40, "IOP Archive Backup-Restore" }
};

unsigned char modulevalue(const char *moduletype)
{
	int i;

	for (i=0; i<sizeof(moduleinfo)/sizeof(struct mod_info); i++)
		if (strcmp(moduleinfo[i].moduletype, moduletype) == 0)
			return moduleinfo[i].value;
	return 0xff;
}

struct i2o_driver *getdst(int fd, int iop)
{
	struct i2o_driver *driver;
	struct i2o_cmd_psetget params;
	u16 opbuf[5];
	unsigned int reslen;
	int oplen;

	driver = (struct i2o_driver *)malloc(sizeof(struct i2o_driver));
	if (!driver) return NULL;

	memset(driver, 0, sizeof(struct i2o_driver));

	opbuf[0] = 1;		/* operation count */
	opbuf[1] = 0;		/* pad */
	opbuf[2] = I2O_PARAMS_TABLE_GET;
	opbuf[3] = 0x0005;	/* DRIVER_STORE_TABLE */
	opbuf[4] = -1;		/* field count */

	oplen  = sizeof(opbuf);
	reslen = sizeof(struct i2o_driver);

	params.iop    = iop;
	params.tid    = I2O_CLASS_EXECUTIVE;
	params.opbuf  = opbuf;
	params.oplen  = oplen;
	params.resbuf = driver;
	params.reslen = &reslen;

	if (ioctl(fd, I2OPARMGET, &params) == 0)
		return driver;
	return NULL;
}

char *modinfo(struct I2O_MODULE_DESC_HDR *hdr)
{
	char *r, *s;
	
	r = (char *)malloc(MAXLEN);
	if (!r) return NULL;
	s = r;

	s += sprintf(s, "Module header size: %d\n", hdr->headerSize);
	s += sprintf(s, "Organization ID   : %04x\n", hdr->orgId);
	s += sprintf(s, "Module ID         : %04x\n", hdr->moduleId);
	s += sprintf(s, "Module date       : %.4s/%.2s/%.2s\n",
	 		(char *)&(hdr->year), (char *)&(hdr->month), (char *)&(hdr->day));
	s += sprintf(s, "I2O version       : %d\n", hdr->i2oVersion);
	s += sprintf(s, "Capabilities      : %02x\n", hdr->majorCapabilities);
	s += sprintf(s, "Code size         : %d\n", hdr->codeSize);
	s += sprintf(s, "Module version    : %.4s\n", hdr->moduleVersion);
	s += sprintf(s, "IOP type          : %d\n", hdr->processorType);
	s += sprintf(s, "IOP version       : %d\n", hdr->processVersion);
	s += sprintf(s, "Module format     : %d\n", hdr->objCodeFormat);
	s += sprintf(s, "Module info       : %.24s\n", hdr->moduleInfo);

	return r;
}             
